MinimalObject subclass: #Future
	instanceVariableNames: 'value semaphore '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Abcl'!
Future comment:
'Class Future implements future objects which hold a value not computed yet.
Future objects are used to implement future type message passing, that is returning eagerly a promise of a value to be computed.
Class Future is very close to class CBox of ConcurrentSmalltalk. (see in the book OOCP).
Class Future is a subclass of class MinimalObject (see below).
A future is implemented with a value and a semaphore to control access to this value (mainly wait until the assignment).

Instance Variables:

	value		<Object>		holds the value. May be not assigned yet.
	semaphore	<Semaphore>	synchronizes accesses to the value
								(accessing processes are suspended until the value is set).

Instance method reply: is the assignment method. We use the selector reply: because in future type message passing a future object becomes the reply destination of the requested computation.
Instance method value fetchs the value of the future and suspends until assignment.
Any other message sent to a future is implicitly (and transparently) forwarded to the value fetched (and thus suspended until assignment). This works because class Future is a subclass of class MinimalObject.
Because the equality message (=) is defined in class MinimalObject, we redefine it in class Future to forward it to the value fetched.'!


!Future methodsFor: 'initialize'!

initialize

	semaphore := Semaphore new! !

!Future methodsFor: 'accessing'!

reply: aValue
	"Set the value of a future."

	value := aValue.
	semaphore signal	"Future is now set/ready."!

value
	"Consult the value of a future. If the value is not yet set wait until ready."

	semaphore wait;	"Ensure that the future is ready."
			signal.	"Reset the semaphore afterwards for next access."
	^value! !

!Future methodsFor: 'implicit access'!

doesNotUnderstand: aMessage
	"Sending a message is transparently delegated to the value of the future.
	The future behaves as the value it holds."

	^self value performMessage: aMessage! !

!Future methodsFor: 'comparing'!

= anObject
	"Transparently force access to the value."

	^self value = anObject! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Future class
	instanceVariableNames: ''!


!Future class methodsFor: 'instance creation'!

new

	^super new initialize! !

Future subclass: #SAFuture
	instanceVariableNames: 'isReady isReadySemaphore '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Abcl'!
SAFuture comment:
'Class SAFuture ensures single assignment of futures.
It is a subclass of class Future.

Instance Variables:

	isReady				<Boolean>		predicate telling if assignment has been already done.
	isReadySemaphore	<Semaphore>	protects assignment of the predicate above.'!


!SAFuture methodsFor: 'initialize'!

initialize

	super initialize.
	isReadySemaphore := Semaphore forMutualExclusion.
	isReadySemaphore critical: [isReady := false]! !

!SAFuture methodsFor: 'accessing'!

reply: aValue
	"Set the value of a future. a SAFuture enforces there is at most one assignment."

	isReadySemaphore critical:
		[isReady
			ifTrue: [self error: 'cannot be assigned twice']
			ifFalse:
				[isReady := true.	"Future is now set/ready."
				super reply: aValue]]! !

!SAFuture methodsFor: 'printing'!

printOn: aStream
	"When value is set, print out as its value."

	isReady
		ifTrue: [value printOn: aStream]
		ifFalse: [super printOn: aStream]! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SAFuture class
	instanceVariableNames: ''!


!SAFuture class methodsFor: 'counter example'!

example
	"SAFuture example"

	"Should raise an error."

	SAFuture new reply: 1; reply: 2! !

Actor subclass: #AbclActor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Abcl'!
AbclActor comment:
'Class AbclActor implements the three Abcl/1 types of message passing:
	now		synchronous
	past		asynchronous, as in standard Actalk
	future	returning eagerly a future object (see class Future)

There exists a special kind of future object (class MAfuture, MA stands for multiple assignments) which accepts multiple returned values as in initial Abcl/1.
On Abcl/1, see the book ABCL/1, edited by Akinori Yonezawa, MIT-Press, 1990.

Warning: recursive synchronous message passing immediately leads to a deadlock.
So don''t send a now message to aself!!

See examples in Actalk-Ext-Abcl-Examples.'!


!AbclActor methodsFor: 'message passing'!

dispatchMessageSend: aMessage on: lastArgument
	"Dispatch along the three types of message passing (now, past, future) of Abcl/1.
	Check last argument to be symbol #now, #future, or #MAfuture.
	Dispatch is done using sequence of tests with ^ escape/return (as for factorial method)."

	lastArgument == #now
		ifTrue: [^self nowSend: aMessage].
	lastArgument == #future
		ifTrue: [^self futureSend: aMessage].
	lastArgument == #MAfuture
		ifTrue: [^self MAfutureSend: aMessage].
	^self asynchronousSend: aMessage!

doesNotUnderstand: aMessage
	"Check last argument to look for non asynchronous transmissions (now or future type).
	In case of now or future type message passing the reply destination slot
	(assumed to be the last argument) is unused and contains symbol #now, #future, or #MAfuture."

	^aMessage arguments isEmpty
		ifTrue: [self asynchronousSend: aMessage]
		ifFalse: [self dispatchMessageSend: aMessage on: aMessage arguments last]!

futureSend: aMessage
	"Implement future type message passing as an asynchronous message send
	with a new future object as a reply destination.
	Immediately return the future object."

	| aFuture |
	aFuture := SAFuture new.
	"Assign the future object as the actual reply destination (the last argument)."
	aMessage arguments at: aMessage arguments size put: aFuture.
	self asynchronousSend: aMessage.
	^aFuture!

MAfutureSend: aMessage
	"Implement future type message passing as an asynchronous message send
	with a new future object as a reply destination (accepting multiple assignments).
	Immediately return the future object."

	| aFuture |
	aFuture := MAFuture new.
	"Assign the future object as the actual reply destination (the last argument)."
	aMessage arguments at: aMessage arguments size put: aFuture.
	self asynchronousSend: aMessage.
	^aFuture!

nowSend: aMessage
	"Implement synchronous message passing as waiting onto a future type message passing."

	^(self futureSend: aMessage) value! !

AbclActor subclass: #ExpressAbclActor
	instanceVariableNames: 'expressMailBox '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Abcl'!
ExpressAbclActor comment:
'Class ExpressAbclActor implements express mode message passing.
Express mode messages take priority over normal mode messages (like an interruption).
They are tagged as starting with a E.

Class ExpressAbclActor is a subclass of class AbclActor.

Instance Variables:

	expressMailBox		<MailBox>	a mailqueue containing express messages.'!


!ExpressAbclActor methodsFor: 'initialize'!

actalkPrivateInitialize
	"Initialize the express mailbox."

	super actalkPrivateInitialize.
	expressMailBox := MailBox new! !

!ExpressAbclActor methodsFor: 'accessing'!

expressMailBox

	^expressMailBox! !

!ExpressAbclActor methodsFor: 'message passing'!

doesNotUnderstand: aMessage
	"Check if this is an express mode message,
	according to the first letter of the selector (E for express).
	To simplify, we assume that express mode may only be in asynchronous type."

	^(aMessage selector at: 1) = $E
		ifTrue: [self expressAsynchronousSend: aMessage]
		ifFalse: [super doesNotUnderstand: aMessage]!

expressAsynchronousSend: aMessage
	"Deposit the message into the express mailBox."

	expressMailBox nextPut: aMessage.
	behavior kernelEventReceive: aMessage! !

ActorBehavior subclass: #AbclActorBehavior
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Abcl'!
AbclActorBehavior comment:
'Class AbclActorBehavior implements the waitFor:andDo: construct, that is explicit acceptance of a message among a set of selectors. Once accepted the argument block is evaluated with the arguments of the message bound to the parameters of the block.
An example of the use of this construct is given in the bounded buffer example in category Actalk-Ext-Abcl-Example.
Three new generic event handling methods are associated to the waitFor construct (when entering a waitFor, when it accepts a message, and for its completion). They are mostly used for tracing purposes.'!


!AbclActorBehavior methodsFor: 'message handling'!

waitFor: arrayOfSelectors andDo: aBlock
	"Look into the mailbox for the first message matching one of the selectors,
	and then accept it.
	It is blocking and suspends until it finds any matching one.
	(This means it suspends onto incoming messages as long as necessary)."
	"Trigger three types of generic events related to waitFor construct."

	| aMessage |
	self eventWaitFor: arrayOfSelectors.		"Generic event triggering."
	aMessage := aself mailBox firstMessageAmong: arrayOfSelectors.
	self eventWaitForAccept: aMessage.		"Generic event triggering."
	aBlock valueWithArguments: aMessage arguments.
	self eventWaitForComplete: aMessage		"Generic event triggering."! !

!AbclActorBehavior methodsFor: 'events'!

eventWaitFor: arrayOfSelectors
	"This method is called when the behavior enters a waitFor construct."
	"This method may be redefined in subclasses."
	"Default behavior is empty."

	"Below are possible examples of use to:
trace:
	Transcript show: self printString , ' waitFor (' , aMessage compactPrintString , ')'; cr.
schedule:
	Processor yield.
broadcast change to dependents (e.g., useful for MVC interface):
	self changed: #eventWaitFor with: aMessage.
step:
	self halt.
	"!

eventWaitForAccept: aMessage
	"This method is called when the behavior start accepting a message within a waitFor construct."
	"This method may be redefined in subclasses."
	"Default behavior is empty."

	"Below are possible examples of use to:
trace:
	Transcript show: self printString , ' waitForAccept (' , aMessage compactPrintString , ')'; cr.
schedule:
	Processor yield.
broadcast change to dependents (e.g., useful for MVC interface):
	self changed: #eventWaitForAccept with: aMessage.
step:
	self halt.
	"!

eventWaitForComplete: aMessage
	"This method is called when the behavior completed (exited) a waitFor construct."
	"This method may be redefined in subclasses."
	"Default behavior is empty."

	"Below are possible examples of use to:
trace:
	Transcript show: self printString , ' waitForComplete (' , aMessage compactPrintString , ')'; cr.
schedule:
	Processor yield.
broadcast change to dependents (e.g., useful for MVC interface):
	self changed: #eventWaitForComplete with: aMessage.
step:
	self halt.
	"! !

!AbclActorBehavior methodsFor: 'actor creation'!

actorClass

	^AbclActor! !

AbclActorBehavior subclass: #ExpressAbclActorBehavior
	instanceVariableNames: 'expressProcess expressExclusionSemaphore '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Abcl'!
ExpressAbclActorBehavior comment:
'Class ExpressAbclActorBehavior computes express mode messages.
A second behavior autonomous process is created for handling express messages. When resumed (when an express message arrived), it suspends standard behavior process while computing an express message.
Method nonResume may be used within an express message method. It will abort resumption of computing a normal message if one was being interrupted.
Method expressAtomic: may be used within any standard method. It ensures that computation won''t be interrupted by a possibly incoming express message.

Class ExpressAbclActorBehavior is a subclass of class AbclActorBehavior.

Instance Variables:

	expressProcess				<Process>		the process for computing express messages.
	expressExclusionSemaphore 	<Semaphore>	ensures mutual exclusion of express messages
												with the expressAtomic: construct.'!


!ExpressAbclActorBehavior methodsFor: 'initialize'!

actalkPrivateInitialize
	"Initialize the semaphore for ensuring atomicity against express message interruption."

	super actalkPrivateInitialize.
	expressExclusionSemaphore := Semaphore forMutualExclusion! !

!ExpressAbclActorBehavior methodsFor: 'process setting'!

setProcess
	"Create a second behavior autonomous process for express messages."

	super setProcess.
	expressProcess := [[true] whileTrue:
						[self acceptExpressMessage: aself expressMailBox next]] newProcess.
	expressProcess resume! !

!ExpressAbclActorBehavior methodsFor: 'process control'!

terminate
	"Terminate also express message behavior process."

	"We have to launch a specific process to terminate
	both behavior process and express behavior process.
	At first look it may seem very paradoxal indeed!!
	However the point is that we don't know which current process is:
	behavior process, express behavior process, or an external process.
	We have to terminate two processes in sequence.
	If the first process we terminate happens to be current process
	we would never terminate the second one!!"

	[super terminate.
	expressProcess terminateProcess] fork! !

!ExpressAbclActorBehavior methodsFor: 'message handling'!

acceptExpressMessage: aMessage
	"Accepting an express message takes priority (suspend) over normal (standard) acceptance.
	Mutually exclude with the expressAtomic: construct."

	expressExclusionSemaphore critical:
		[process suspend.
		self acceptMessage: aMessage.
		process resume]!

expressAtomic: aBlock
	"Ensure that computation expressed within aBlock won't be interrupted by an express message.
	May be used within script methods as well as private routines."

	expressExclusionSemaphore critical: aBlock!

nonResume
	"Terminate current standard behavior process and restart a new one.
	In case a normal message was being processed, it will be aborted."

	"Should be used ONLY by express message methods!!
	(Otherwise if used within standad methods,
	it would abort main behavior process which is current process!!)."

	process terminate.
	process := self behaviorProcessBlock newProcess! !

!ExpressAbclActorBehavior methodsFor: 'actor creation'!

actorClass

	^ExpressAbclActor! !

SharedQueue subclass: #MAFuture
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Abcl'!
MAFuture comment:
'Class MAFuture (multiple assignment future) implements future objects which may hold several values (as in initial Abcl/1 model).
Class SharedQueue implements such basic functionalities. We just need to rename the accessing methods.'!


!MAFuture methodsFor: 'accessing'!

reply: aValue
	"Set the value of a future. For a MAFuture this means enqueueing the value."

	self nextPut: aValue!

value
	"Consult the value of a future. If the value is not yet set wait until ready."

	^self peek! !

!MAFuture methodsFor: 'printing'!

printOn: aStream
	"Print out as a standard SharedQueue (that is as its contents: an ordered collection),
	bur prefixed with the tag of a MAFuture."

	aStream nextPutAll: self class name , ':'.
	super printOn: aStream! !
