IRActorBehavior subclass: #QuickSort
	instanceVariableNames: 'leftSort pivot rightSort '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Accept-Examples'!
QuickSort comment:
'Class QuickSort implements a distributed sorter.
This simple algorithm has been initially described in ABCL/1 by Etsuya Shibayama and Akinori Yonezawa in Distributed Computing in ABCL/1, in the book OOCP, edited by A. Yonezawa and M. Tokoro, MIT-Press, 1987, pages 123-126. This divide and conquer algorithm regards the data as active objects and requests as messages. This makes the synchronization easy as opposed if we would let concurrent processes access a passive tree.
The initial algorithm has been someway transformed and simplified here.

Each QuickSort tree is recursively decomposed into a pivot value and two subtrees, left and right, according to elements lower or greater than the pivot.
Note that the method insert:andAckTo: specifies a reply destination to which an acknowledgement message (oneInserted) will be issued. The client (behavior QuickSortClient) will accept (wait for) as many oneInserted messages as there are elements to be sorted. Therefore the client will be able to know when the sorting is completed.

Class QuickSort is a subclass of class IRActorBehavior, as it uses synchronous message passing to recursively retrieve elements.
See also class QuickSortClient comment.

See also a similar example of a distributed symbol table in category Actalk-Ext-Pool-Examples.'!


!QuickSort methodsFor: 'script'!

insert: value andAckTo: r
	"Insert an element and acknowledge to the reply destination (client)."

	pivot isNil
		ifTrue:
			[pivot := value.
			r oneInserted]	"Acknowledge that one element has been inserted."
		ifFalse:
			[value < pivot
				ifTrue:
					[leftSort isNil
						ifTrue: [leftSort := self class new actor].
					leftSort insert: value andAckTo: r]
				ifFalse:
					[rightSort isNil
						ifTrue: [rightSort := self class new actor].
					rightSort insert: value andAckTo: r]]!

retrieveAndAddTo: aCollection
	"Retrieve the sorted elements by adding them to a collection."

	leftSort isNil
		ifFalse: [(leftSort retrieveAndAddTo: aCollection) value].
	aCollection add: pivot.
	rightSort isNil
		ifFalse: [(rightSort retrieveAndAddTo: aCollection) value]! !

AcceptActorBehavior subclass: #QuickSortClient
	instanceVariableNames: 'sort n '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Accept-Examples'!
QuickSortClient comment:
'ClassQuickSortClient is used to control a distributed quick sort (behavior QuickSort).
Class QuickSortClient is defined as a subclass of class AcceptActorBehavior, as it uses the answer: construct to determine when sorting is completed (by accepting n messages oneInserted acknowledged by the QuickSort tree).
In order to display the result in a comprehensive form, it retrieves the elements once sorted in the QuickSort distributed tree, and recollects them into a collection.

See class QuickSort comment.'!


!QuickSortClient methodsFor: 'script'!

oneInserted
	"Do nothing.
	Used for waiting for n elements to be inserted by accepting n messages oneInserted."!

retrieve
	"Retrieve the elements sorted into a new collection and return it."

	| aCollection |
	aCollection := OrderedCollection new.
	(sort retrieveAndAddTo: aCollection) value.
	Transcript show: 'Result: ' , aCollection printString; cr.
	^aCollection!

sort: aCollection
	"Sort the elements within the collection. Retrieve the result into a new collection."

	Transcript show: 'Sort: ' , aCollection printString; cr.
	sort := QuickSort new actor.
	n := aCollection size.
	aCollection do: [:value | sort insert: value andAckTo: aself].	"Insert all elements."
	n timesRepeat:					"Wait for completion of insertion/sorting."
		[self answer: #(oneInserted)].
	Transcript show: 'Sort done.'; cr.
	^self retrieve! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

QuickSortClient class
	instanceVariableNames: ''!


!QuickSortClient class methodsFor: 'example'!

example
	"QuickSortClient example inspect"

	^(QuickSortClient new actor
		sort: #(3 7 5 1 8 2 10 6 9 4)) value!

example: size
	"QuickSortClient example: 100"

	"Sort a random collection of a certain size."

	| aCollection random|
	aCollection := OrderedCollection new.
	random := Random new.
	size timesRepeat:
		[aCollection add: (random next * size) rounded].
	QuickSortClient new actor
		sort: aCollection! !

AcceptActorBehavior subclass: #AcceptBoundedBuffer
	instanceVariableNames: 'contents maxSize '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Accept-Examples'!
AcceptBoundedBuffer comment:
'Class AcceptBoundedBuffer implements the bounded buffer example.
If a buffer accepted a disabled request (e.g., a put: message once it is full), it will explicitly wait for an alternative request (e.g., a get request) in order to first free the buffer.

See definition of put: and get methods.
Note that as for class AbclBoundedBuffer, synchronization code appears explicitly within the method body. As a result this model is not very good at reusing specifications.
See discussion of inheritance anomalies in examples of categories Actalk-Synchro-*.'!


!AcceptBoundedBuffer methodsFor: 'initialize'!

initialize: anInteger

	maxSize := anInteger.
	contents := BoundedBufferObject new maxSize: maxSize! !

!AcceptBoundedBuffer methodsFor: 'script'!

get

	contents isEmpty					"If empty,"
		ifTrue: [self answer: #(put:)].	"then, accept a put:."
	^contents removeFirst				"Now proceed as usual."!

put: item

	contents isFull					"If full,"
		ifTrue: [self answer: #(get)].	"then, accept a get."
	contents addLast: item			"Now proceed as usual."! !

!AcceptBoundedBuffer methodsFor: 'printing'!

printOn: aStream
	"Print as its class name, then its contents."

	super printOn: aStream.
	contents printOn: aStream! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

AcceptBoundedBuffer class
	instanceVariableNames: ''!


!AcceptBoundedBuffer class methodsFor: 'instance creation'!

new: size

	^self new initialize: size! !

!AcceptBoundedBuffer class methodsFor: 'example'!

exampleSize: maxSize numberItems: numberItems speedRatio: ratio
	"AcceptBoundedBuffer exampleSize: 3 numberItems: 8 speedRatio: 4"
	"AcceptBoundedBuffer exampleSize: 3 numberItems: 8 speedRatio: 1/4"

	| buffer producer consumer |
	buffer := (AcceptBoundedBuffer new: maxSize) actor.
	producer := (Producer new buffer: buffer delay: 1) actor.
	consumer := (Consumer new buffer: buffer delay: ratio) actor.
	producer runPut: numberItems.
	consumer runGet: numberItems! !
