AcceptReplaceActorBehavior subclass: #AcceptNullBoundedBuffer
	instanceVariableNames: 'contents maxSize '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Replace-Ext-Ex'!
AcceptNullBoundedBuffer comment:
'Class AcceptNullBoundedBuffer is an improved version of class NullBoundedBuffer (implementing the bounded buffer example).
It relies on controlled acceptance of existing methods as provided by class AcceptReplaceActorBehavior.
The only difference (and improvement) over class NullBoundedBuffer is the disparition of redefinition of undefined message handling to resend the initial message.
Note that thanks to this control over message acceptance, we recover the assumption of transmission message ordering (as opposed to class NullBoundedBuffer).

It is defined as a subclass of class AcceptReplaceActorBehavior.

Instance Variables:

	contents		<BoundedBufferObject>	the bounded buffer object.
	maxSize	<Integer>				its maximum size, that is the maximum number of items.'!


!AcceptNullBoundedBuffer methodsFor: 'initialize'!

contents: aBoundedBufferObject maxSize: anInteger

	contents := aBoundedBufferObject.
	maxSize := anInteger! !

!AcceptNullBoundedBuffer methodsFor: 'printing'!

printOn: aStream
	"Print as its class name, then its contents."

	super printOn: aStream.
	contents printOn: aStream! !

AcceptNullBoundedBuffer subclass: #AcceptEmptyBoundedBuffer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Replace-Ext-Ex'!
AcceptEmptyBoundedBuffer comment:
'Classes AcceptEmptyBoundedBuffer, AcceptPartialBoundedBuffer and AcceptFullBoundedBuffer define the three types of behaviors of a bounded buffer. They follow model of behavior replacement defined by class AcceptReplaceActorBehavior, that is with control over acceptance of messages in order to only accept defined (that is enabled) methods.
As a consequence transmission message ordering is preserved.
A part this difference they are similar to classes EmptyBoundedBuffer, PartialBoundedBuffer and FullBoundedBuffer, see their respective comments in category Actalk-Ext-Replace-Examples.'!


!AcceptEmptyBoundedBuffer methodsFor: 'script'!

put: item

	contents addLast: item.
	self replace: (contents isFull		"Compute replacement behavior."
					ifTrue: [AcceptFullBoundedBuffer new contents: contents maxSize: maxSize]
					ifFalse: [AcceptPartialBoundedBuffer new contents: contents maxSize: maxSize])! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

AcceptEmptyBoundedBuffer class
	instanceVariableNames: ''!


!AcceptEmptyBoundedBuffer class methodsFor: 'instance creation'!

new: maxSize
	"Create a new empty bounded buffer behavior.
	If maxSize is null, create a null bounded buffer behavior."

	^(maxSize > 0
		ifFalse: [AcceptNullBoundedBuffer]		"Bounded buffer of null size."
		ifTrue: [self]) new
			contents: (BoundedBufferObject new maxSize: maxSize)
			maxSize: maxSize! !

!AcceptEmptyBoundedBuffer class methodsFor: 'example'!

exampleSize: maxSize numberItems: numberItems speedRatio: ratio
	"AcceptEmptyBoundedBuffer exampleSize: 3 numberItems: 8 speedRatio: 4"
	"AcceptEmptyBoundedBuffer exampleSize: 3 numberItems: 8 speedRatio: 1/4"

	| buffer producer consumer |
	buffer := (AcceptEmptyBoundedBuffer new: maxSize) actor.
	producer := (Producer new buffer: buffer delay: 1) actor.
	consumer := (Consumer new buffer: buffer delay: ratio) actor.
	producer runPut: numberItems.
	consumer runGet: numberItems replyTo: ConsumerPrinter new actor! !

AcceptEmptyBoundedBuffer subclass: #AcceptPartialBoundedBuffer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Replace-Ext-Ex'!
AcceptPartialBoundedBuffer comment:
'Classes AcceptEmptyBoundedBuffer, AcceptPartialBoundedBuffer and AcceptFullBoundedBuffer define the three types of behaviors of a bounded buffer. They follow model of behavior replacement defined by class AcceptReplaceActorBehavior, that is with control over acceptance of messages in order to only accept defined (that is enabled) methods.
As a consequence transmission message ordering is preserved.
A part this difference they are similar to classes EmptyBoundedBuffer, PartialBoundedBuffer and FullBoundedBuffer, see their respective comments in category Actalk-Ext-Replace-Examples.'!


!AcceptPartialBoundedBuffer methodsFor: 'script'!

getAndReplyTo: r

	r reply: contents removeFirst.
	self replace: (contents isEmpty		"Compute replacement behavior."
					ifTrue: [AcceptEmptyBoundedBuffer new contents: contents maxSize: maxSize]
					ifFalse: [AcceptPartialBoundedBuffer new contents: contents maxSize: maxSize])! !

AcceptNullBoundedBuffer subclass: #AcceptFullBoundedBuffer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Replace-Ext-Ex'!
AcceptFullBoundedBuffer comment:
'Classes AcceptEmptyBoundedBuffer, AcceptPartialBoundedBuffer and AcceptFullBoundedBuffer define the three types of behaviors of a bounded buffer. They follow model of behavior replacement defined by class AcceptReplaceActorBehavior, that is with control over acceptance of messages in order to only accept defined (that is enabled) methods.
As a consequence transmission message ordering is preserved.
A part this difference they are similar to classes EmptyBoundedBuffer, PartialBoundedBuffer and FullBoundedBuffer, see their respective comments in category Actalk-Ext-Replace-Examples.'!


!AcceptFullBoundedBuffer methodsFor: 'script'!

getAndReplyTo: r

	r reply: contents removeFirst.
	self replace: (contents isEmpty		"Compute replacement behavior."
					ifTrue: [AcceptEmptyBoundedBuffer new contents: contents maxSize: maxSize]
					ifFalse: [AcceptPartialBoundedBuffer new contents: contents maxSize: maxSize])! !

ReplaceActorBehavior subclass: #OverdraftManager
	instanceVariableNames: 'account balance savings r '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Replace-Ext-Ex'!
OverdraftManager comment:
'Class OverdraftManager implements overdraft management of a ProtectedBankAccount.
It will compute and send/assign the replacement behavior of the protected bank account.

Instance Variables:

	account		<ProtectedBankAccount actor>	is insensitive during the overdraft management.
	balance		<Integer>						the current balance of the ProtectedBankAccount.
	savings		<BankAccount actor>				the savings account.
	r			<Actor>							the reply destination to which the receipt
												acknowledgement has to be sent.'!


!OverdraftManager methodsFor: 'initialize'!

account: aPBankAccount balance: anInteger savings: aBankAccount r: aReplyDestination

	account := aPBankAccount.
	balance := anInteger.
	savings := aBankAccount.
	r := aReplyDestination! !

!OverdraftManager methodsFor: 'script'!

reply: receipt

	self replace: NoMoreMessage new.	"Used only once."
	(Delay forMilliseconds: 5) wait.		"Simulation of bank network access..."
	receipt = #withdrawalReceipt		"If the savings covers the overdraft,"
		ifTrue:
			[account replace:			"then the protected bank account new balance becomes 0."
				(ProtectedBankAccount new balance: 0 savings: savings).
			r reply: #overdraftCoveredReceipt]			"Acknowledgement."
		ifFalse:
			[account replace:			"Otherwise, we leave it as it is,"
				(ProtectedBankAccount new balance: balance savings: savings).
			r reply: #overdraftNotCoveredReceipt]		"Acknowledgement."! !

BankAccount subclass: #ProtectedBankAccount
	instanceVariableNames: 'savings '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Replace-Ext-Ex'!
ProtectedBankAccount comment:
'Class ProtectedBankAccount implements a bank account with provision and overdraft management.
It is a subclass of class BankAccount.
In order to manage overdrafts, a protected bank account will own some private bank account, called its savings, as an overdraft protection.
In case of overdraft, the protected bank account will inquire its savings for protection, by sending to it a request for withdrawal of the overdraft amount.
If the overdraft could be covered by the savings account, then the overdraft amount is withdrawn from the savings
and used to rebalance the protected bank account (which then becomes null).
If the overdraft is too important for the savings, then previous behavior is unchanged (and a justice decision is inquired!!?!!).

The point is following: the protected bank account cannot decide its further balance (and replacement behavior) before getting the reply (acceptance or reject) from the savings account. Meanwhile it becomes insensitive. A specific continuation will be created to manage the acknowledgement (receipt) from the savings and to compute and send the behavior replacement accordingly. The corresponding class is named OverdraftManager.

Instance Variables:

	savings		<aBankAccount actor>	the savings account.'!


!ProtectedBankAccount methodsFor: 'initialize'!

balance: anInteger savings: aBankAccount

	balance := anInteger.
	savings := aBankAccount! !

!ProtectedBankAccount methodsFor: 'script'!

deposit: amount replyTo: r

	self replace: (self class new balance: balance + amount savings: savings).
	r reply: #depositReceipt!

withdraw: amount replyTo: r
	"In case of overdraft, become insensitive while waiting for management
	(the savings account is inquired for possible covering of the overdraft)."

	amount < balance
		ifTrue:		"Safe case."
			[self replace:
				(self class new balance: balance - amount savings: savings).
			r reply: #withdrawalReceipt]
		ifFalse:		"Overdraft."
			[savings withdraw: (amount - balance)	"Withdrawal request is sent to the savings account"
					replyTo:						"with an overdraft manager as continuation."
						(OverdraftManager new		
									account: aself	"Ths continuation stores all necessary info."
									balance: balance
									savings: savings
									r: r) actor]

	"Note that there is no behavior replacement in case of overdraft.
	The protected bank account actor becomes insensitive until reception of
	a behavior replacement message conveying the replacement behavior computed
	and issued by the overdraft manager continuation.
	It does not reply a receipt either, the overdraft manager will."! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

ProtectedBankAccount class
	instanceVariableNames: ''!


!ProtectedBankAccount class methodsFor: 'instance creation'!

newBalance: balance savingsAmount: savingsAmount
	"Create a bank account and its savings account."

	^self new balance: balance savings: (BankAccount new balance: savingsAmount) actor! !

!ProtectedBankAccount class methodsFor: 'example'!

example
	"ProtectedBankAccount example"

	(ProtectedBankAccount newBalance: 200 savingsAmount: 100) actor
		deposit: 150 replyTo: Print;
		withdraw: 400 replyTo: Print;		"Create an overdraft (50)."
		balanceAndReplyTo: Print! !
